/*
 * $Id: ft_http.c,v 1.2 2003/07/09 04:23:54 jasta Exp $
 *
 * Copyright (C) 2001-2003 giFT project (gift.sourceforge.net)
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2, or (at your option) any
 * later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 */

#include "ft_openft.h"

#include "ft_http.h"

#include <ctype.h>

/*****************************************************************************/

char *http_code_string (int code)
{
	char *str;

	/* we are only going to support the subset that OpenFT uses for now */
	switch (code)
	{
	 case 200:     str = "OK";                        break;
	 case 206:     str = "Partial Content";           break;
	 case 403:     str = "Forbidden";                 break;
	 case 404:     str = "Not Found";                 break;
	 case 500:     str = "Internal Server Error";     break;
	 case 501:     str = "Not Implemented";           break;
	 case 503:     str = "Service Unavailable";       break;
	 default:      str = NULL;                        break;
	}

	if (!str)
	{
		FT->DBGFN (FT, "unknown http code %i", code);
		str = "Unknown";
	}

	return str;
}

/*****************************************************************************/

BOOL http_check_sentinel (char *data, size_t len)
{
	int cnt;

	assert (len > 0);
	len--;

	for (cnt = 0; len > 0 && cnt < 2; cnt++)
	{
		if (data[len--] != '\n')
			break;

		/* treat CRLF as LF */
		if (data[len] == '\r')
			len--;
	}

	return (cnt == 2);
}

/*****************************************************************************/

static int oct_value_from_hex (char hex_char)
{
	if (!isxdigit (hex_char))
		return 0;

	if (hex_char >= '0' && hex_char <= '9')
		return (hex_char - '0');

	hex_char = toupper (hex_char);

	return ((hex_char - 'A') + 10);
}

char *http_url_decode (const char *encoded)
{
	char *decoded, *ptr;

	if (!encoded)
		return NULL;

	/* make sure we are using our own memory here ... */
	ptr = strdup (encoded);

	/* save the head */
	decoded = ptr;

	/* convert '+' -> ' ' and %2x -> char value */
	while (*ptr)
	{
		switch (*ptr)
		{
		 case '+':
			*ptr = ' ';
			break;
		 case '%':
			if (isxdigit (ptr[1]) && isxdigit (ptr[2]))
			{
				int oct_val;

				oct_val =  oct_value_from_hex (ptr[1]) * 16;
				oct_val += oct_value_from_hex (ptr[2]);

				*ptr = (char) oct_val;

				string_move (ptr + 1, ptr + 3);
			}
			break;
		 default:
			break;
		}

		ptr++;
	}

	return decoded;
}

static char *url_encode_char (char *stream, unsigned char c)
{
	*stream++ = '%';

	sprintf (stream, "%02x", (unsigned int) c);

	return stream + 2;
}

char *http_url_encode (const char *decoded)
{
	char *encoded, *ptr;

	if (!decoded)
		return NULL;

	/* allocate a large enough buffer for the largest possible case */
	encoded = ptr = malloc ((strlen (decoded) * 3) + 1);

	while (*decoded)
	{
		/* we can rule out non-printable and whitespace characters */
		if (!isprint (*decoded) || isspace (*decoded))
			ptr = url_encode_char (ptr, *decoded);
		else
		{
			/* check for anything special */
			switch (*decoded)
			{
			 case '?':
			 case '#':
			 case '@':
			 case '+':
			 case '%':
			 case '&':
			 case ':':
			 case '=':
			 case '(':
			 case ')':
			 case '[':
			 case ']':
			 case '\"':
			 case '\'':
				ptr = url_encode_char (ptr, *decoded);
				break;
			 default: /* regular character, just copy */
				*ptr++ = *decoded;
				break;
			}
		}

		decoded++;
	}

	*ptr = 0;

	return encoded;
}

/*****************************************************************************/

static unsigned int http_parse_keylist (Dataset **d, char *data)
{
	char *row;
	char *key;
	char *value;

	if (!d || !data)
		return 0;

	while ((row = string_sep_set (&data, "\r\n")))
	{
		if (!(key = string_sep (&row, ": ")) || !row)
			continue;

		value = row;

		dataset_insertstr (d, key, value);
	}

	return dataset_length (*d);
}

/*****************************************************************************/

FTHttpReply *ft_http_reply_new (int code)
{
	FTHttpReply *reply;

	if (!(reply = MALLOC (sizeof (FTHttpReply))))
		return NULL;

	reply->version = 1.0;              /* HTTP/1.0 by default */
	reply->keylist = dataset_new (DATASET_LIST);
	reply->code    = code;

	return reply;
}

void ft_http_reply_free (FTHttpReply *reply)
{
	if (!reply)
		return;

	dataset_clear (reply->keylist);
	free (reply);
}

FTHttpReply *ft_http_reply_unserialize (char *data)
{
	FTHttpReply *reply;
	char        *resp;
	int          vermaj;
	int          vermin;
	int          code;

	assert (data != NULL);

	/* should be something like: HTTP/1.0 200 OK\r\n..., we want just that
	 * first line here */
	if (!(resp = string_sep_set (&data, "\r\n")))
		return NULL;

	/* examine the first line (resp) closer */
	                       string_sep (&resp, "/");   /* shift past HTTP/ */
	vermaj = gift_strtoul (string_sep (&resp, "."));  /* shift past 1. */
	vermin = gift_strtoul (string_sep (&resp, " "));  /* shift past 0  */
	code   = gift_strtoul (string_sep (&resp, " "));  /* shift past 200 */

	if (!(reply = ft_http_reply_new (code)))
		return NULL;

	/* override the default version of 1.0 */
	reply->version = ((float)vermaj) + ((float)vermin / 10.0);

	/* parse the rest (the key/value list) */
	http_parse_keylist (&reply->keylist, data);

	return reply;
}

static void add_keylist (ds_data_t *key, ds_data_t *value, String *str)
{
	string_appendf (str, "%s: %s\r\n", key->data, value->data);
}

char *ft_http_reply_serialize (FTHttpReply *reply, size_t *retlen)
{
	String     *str;
	const char *codestr;

	if (!reply)
		return NULL;

	codestr = http_code_string (reply->code);
	assert (codestr != NULL);

	if (!(str = string_new (NULL, 0, 0, TRUE)))
		return NULL;

	string_appendf (str, "HTTP/%.01f %i %s\r\n",
	                reply->version, reply->code, codestr);
	dataset_foreach (reply->keylist, DS_FOREACH(add_keylist), str);
	string_appendf (str, "\r\n");

	if (retlen)
		*retlen = str->len;

	return string_free_keep (str);
}

int ft_http_reply_send (FTHttpReply *reply, TCPC *c)
{
	char  *data;
	size_t len;
	int    ret;

	data = ft_http_reply_serialize (reply, &len);

	/* as promised, we will handle memory management */
	ft_http_reply_free (reply);

	if (!data)
		return -1;

	ret = tcp_write (c, (unsigned char *)data, len);
	free (data);

	return ret;
}

/*****************************************************************************/

FTHttpRequest *ft_http_request_new (const char *method, const char *request)
{
	FTHttpRequest *req;

	if (!(req = MALLOC (sizeof (FTHttpRequest))))
		return NULL;

	req->method  = STRDUP (method);
	req->request = STRDUP (request);
	req->keylist = dataset_new (DATASET_LIST);

	return req;
}

void ft_http_request_free (FTHttpRequest *req)
{
	if (!req)
		return;

	dataset_clear (req->keylist);
	free (req->method);
	free (req->request);
	free (req);
}

FTHttpRequest *ft_http_request_unserialize (char *data)
{
	FTHttpRequest *req;
	char          *resp;
	char          *method;
	char          *request;

	assert (data != NULL);

	/* data should begin with METHOD requesturi HTTP/version\r\n */
	if (!(resp = string_sep_set (&data, "\r\n")))
		return NULL;

	method  = string_sep (&resp, " ");    /* METHOD */
	request = string_sep (&resp, " ");    /* requesturi */
	                                      /* [ version is not important ] */

	if (!(req = ft_http_request_new (method, request)))
		return NULL;

	http_parse_keylist (&req->keylist, data);

	return req;
}

char *ft_http_request_serialize (FTHttpRequest *req, size_t *retlen)
{
	String *str;

	if (!req)
		return NULL;

	if (!(str = string_new (NULL, 0, 0, TRUE)))
		return NULL;

	string_appendf (str, "%s %s HTTP/1.0\r\n", req->method, req->request);
	dataset_foreach (req->keylist, DS_FOREACH(add_keylist), str);
	string_appendf (str, "\r\n");

	if (retlen)
		*retlen = str->len;

	return string_free_keep (str);
}

int ft_http_request_send (FTHttpRequest *req, TCPC *c)
{
	char  *data;
	size_t len;
	int    ret;

	data = ft_http_request_serialize (req, &len);

	/* as promised, we will handle memory management */
	ft_http_request_free (req);

	if (!data)
		return -1;

	ret = tcp_write (c, (unsigned char *)data, len);
	free (data);

	return ret;
}
